// The MIT License

// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Author: Kermin Fleming kfleming@mit.edu
// De-AWBized by: Richard Uhler ruhler@mit.edu

import Connectable::*;
import GetPut::*;
import ClientServer::*;
import FIFO::*;
import Vector::*;
import Complex::*;
import FixedPoint::*;
import FShow::*;

import AudioProcessorTypes::*;
import FIRFilter::*;
import FFT::*;
import marshaller::*;
import demarshaller::*;
import Padder::*;
import Depadder::*;

`define FFT_POINTS 8


module mkAudioPipeline (AudioProcessor);
  // Pipeline processing elements
  AudioProcessor filter <- mkFIRFilter();
  let fft <- mkFFT();  

  // We need a marshaller/demarshaller to expand and contract the wider fft
  // input type.
  DEMARSHALLER#(Complex#(FixedPoint#(16,16)), Vector#(`FFT_POINTS, Complex#(FixedPoint#(16,16)))) demarshaller  <- mkDeMarshaller();
  MARSHALLER#(Vector#(`FFT_POINTS, Complex#(FixedPoint#(16,16))), Complex#(FixedPoint#(16,16))) marshaller  <- mkMarshaller();
  Padder#(`FFT_POINTS) padder <- mkPadder();
  Depadder#(`FFT_POINTS) depadder <- mkDepadder();

  // Local state, mostly needed for padding and keeping track of EOF 
  FIFO#(Bool) endOfFile <- mkSizedFIFO(2 * `FFT_POINTS); 
  FIFO#(AudioProcessorUnit) outfifo <- mkFIFO();
  FIFO#(Bool) fftOutputsFrequency <- mkSizedFIFO(1);
  FIFO#(Bool) fftOutputsTime      <- mkSizedFIFO(1);

  function data getRel(Complex#(data) val); 
    return val.rel;
  endfunction

  // Due to padding, the FFT will always deal with FFT_POINTS
  rule startDemarshaller (True);
    demarshaller.start(`FFT_POINTS);
  endrule

  rule filter2padder (True);
    let x <- filter.getSampleOutput();
    padder.pipeline.putSampleInput(x);
  endrule

  mkConnection(padder.padding,depadder.padding);

  rule padderToDemarshaller (True);
    let padderData <- padder.pipeline.getSampleOutput();

    case (padderData) matches
      tagged Sample .sample:
        begin 
          endOfFile.enq(False); 
          demarshaller.insert(cmplx(fromInt(sample) * fromReal(1 / `FFT_POINTS), 0));
        end
      tagged EndOfFile:
        begin
          endOfFile.enq(True);
        end  
    endcase       
  endrule

  rule demarshallerToFFT (True);
    let demarshalledData <- demarshaller.readAndDelete();
    fft.fftInput(demarshalledData);
    fftOutputsFrequency.enq(?);
  endrule

  rule fftToIFFT (True);
    let fftData <- fft.fftOutput();
    fft.fftInput(fftData);
    fftOutputsFrequency.deq();
    fftOutputsTime.enq(?);
  endrule
  
  rule fftToMarshaller (True);
    let ifftData <- fft.fftOutput();
    fftOutputsTime.deq();
    marshaller.enq(append(tail(ifftData), cons(head(ifftData), nil)), `FFT_POINTS);
  endrule

  rule marshallerToOutData (!endOfFile.first());
    marshaller.deq();
    depadder.pipeline.putSampleInput(tagged Sample fxptGetInt(marshaller.first().rel));
    endOfFile.deq();
  endrule

  rule marshallerToOutEndOfFile (endOfFile.first());
    depadder.pipeline.putSampleInput(tagged EndOfFile);
    endOfFile.deq();
  endrule
 
  
  method Action putSampleInput(AudioProcessorUnit in) = filter.putSampleInput(in);
  method ActionValue#(AudioProcessorUnit) getSampleOutput() = depadder.pipeline.getSampleOutput();

endmodule
