// The MIT License

// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Author: Kermin Fleming kfleming@mit.edu
// De-AWBized: Richard Uhler ruhler@mit.edu

import FIFO::*;
import FIFOF::*;
import GetPut::*;

import AudioProcessorTypes::*;

interface Padder#(numeric type blockSize);
  interface AudioProcessor pipeline;
  interface Get#(Bit#(TLog#(blockSize))) padding;
endinterface

module mkPadder(Padder#(blockSize));

  FIFO#(AudioProcessorUnit) inFifo <- mkFIFO();
  FIFO#(AudioProcessorUnit) outFifo <- mkFIFO();  
  FIFO#(Bit#(TLog#(blockSize))) paddingFifo <- mkFIFO();
  Reg#(Bit#(TLog#(blockSize))) paddingCount <- mkReg(0);
  Reg#(Bool) insertPadding <- mkReg(False);


  rule noPadRule(!insertPadding);
    inFifo.deq(); 
    case(inFifo.first()) matches
      tagged EndOfFile: 
        begin
          $display("PadderGotEndOfFile");
          insertPadding <= True;
          paddingFifo.enq(paddingCount);
        end
    // Depending on the value of data, we may need to pad out the demarshaller;
      tagged Sample .sample:
        begin
          paddingCount <= paddingCount - 1; // Rely on the underflow of this variable
          outFifo.enq(inFifo.first);
          if(paddingCount - 1 == 0)
            begin
              paddingFifo.enq(0);
            end
        end 
    endcase
  endrule

  // We have to pad out 
  rule padRule (insertPadding);
    if(paddingCount == 0) 
      begin
        $display("PadderEmitsEndOfFile");
        outFifo.enq(tagged EndOfFile);
        insertPadding <= False;
      end 
    else
      begin 
        outFifo.enq(tagged Sample 0);
        paddingCount <= paddingCount - 1;
      end
  endrule


  interface AudioProcessor pipeline;
    method Action putSampleInput(AudioProcessorUnit in) = inFifo.enq(in);
    method ActionValue#(AudioProcessorUnit) getSampleOutput();
        outFifo.deq();
        return outFifo.first();
    endmethod
  endinterface
  
  interface padding = toGet(paddingFifo);

endmodule

