import Complex::*;
import Vector::*;
import FixedPoint::*;
import FIFO::*;
import Real::*;

`define NUM_POINTS 8
`define LOG_NUM_POINTS 3
typedef FixedPoint#(16, 16) COMPLEX_DATA;
typedef Complex#(COMPLEX_DATA) FFT_DATA;

interface FFT;
  method Action fftInput(Vector#(`NUM_POINTS, FFT_DATA) inVector);
  method ActionValue#(Vector#(`NUM_POINTS, FFT_DATA)) fftOutput();
endinterface

module mkFFT (FFT);

  FIFO#(Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA))) inputFIFO  <- mkFIFO(); 
  FIFO#(Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA))) outputFIFO <- mkFIFO(); 

  // Compute the twiddle factor.
  function Complex#(COMPLEX_DATA)  twiddle(Integer i);
     return cmplx(fromReal(cos(fromInteger(i)*pi/`NUM_POINTS)),
                  fromReal(-1*sin(fromInteger(i)*pi/`NUM_POINTS)));
  endfunction

  function Complex#(COMPLEX_DATA) getTwiddle(Integer stage, Integer index);
    return twiddle(((2*index)/(2 ** (`LOG_NUM_POINTS-stage))) * 
	                             (2 ** (`LOG_NUM_POINTS-stage)));
  endfunction

  // Construct the permutation table.
  Vector#(`NUM_POINTS, Integer) permute = newVector();
  for(Integer i = 0; i < `NUM_POINTS/2; i=i+1)
    begin 
      permute[i] = i*2; 
      permute[i + `NUM_POINTS/2] = i*2+1;
    end
  
  // this function reorders the vector by swapping words at positions
  // corresponding to the bit-reversal of their indices.
  // The reordering can be done either as as the
  // first or last phase of the transformation.
  function Vector#(`NUM_POINTS,Complex#(COMPLEX_DATA)) bitReverse(Vector#(`NUM_POINTS,Complex#(COMPLEX_DATA)) inVector);
    Vector#(`NUM_POINTS,Complex#(COMPLEX_DATA)) outVector = newVector();
    for(Integer i = 0; i < `NUM_POINTS; i = i+1)
      begin   
        Bit#(`LOG_NUM_POINTS) reversal = reverseBits(fromInteger(i));
        outVector[reversal] = inVector[i];           
      end  
    return outVector;
  endfunction

  // 2-way Butterfly
  function Vector#(2, Complex#(COMPLEX_DATA)) bfly2(Vector#(2, Complex#(COMPLEX_DATA)) t, Complex#(COMPLEX_DATA) k);
    Complex#(COMPLEX_DATA) m = t[1] * k;

    Vector#(2, Complex#(COMPLEX_DATA)) z = newVector();
    z[0] = t[0] + m;
    z[1] = t[0] - m; 

    return z;
  endfunction

  // This function implements a single stage of the FFT, performing
  // butterflys and a single permutation.
  function Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA)) stage_f(Integer stage, Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA)) stage_in);
    Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA)) stage_temp = newVector();
    for(Integer i = 0; i < (`NUM_POINTS/2); i = i+1)
      begin    
        Integer idx = i * 2;
        let twid = getTwiddle(stage, i);
        let y = bfly2(takeAt(idx, stage_in), twid);

        stage_temp[idx]   = y[0];
        stage_temp[idx+1] = y[1];
      end 

    Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA)) stage_out = newVector();
    for (Integer i = 0; i < `NUM_POINTS; i = i+1)
      begin
        stage_out[i] = stage_temp[permute[i]];
      end
    return stage_out;
  endfunction

  // fft function.
  function Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA)) fft(Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA)) in_data);
  
    Vector#(TAdd#(1, `LOG_NUM_POINTS), Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA))) stage_data = newVector();
    stage_data[0] = in_data;

    for(Integer stage = 0; stage < `LOG_NUM_POINTS; stage=stage+1) 
      begin
        stage_data[stage+1] = stage_f(stage, stage_data[stage]);  
      end
     return stage_data[`LOG_NUM_POINTS];
  endfunction

  // This rule performs fft using a big mass of combinational logic.
  rule comb_fft;
    inputFIFO.deq();
    outputFIFO.enq(fft(inputFIFO.first()));
  endrule
 
  // Interface method used to input Data to the FFT module.  Examine
  // AudioPipeline.bsv to see how this interface method is used
  method Action fftInput(Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA)) inVector);
    inputFIFO.enq(bitReverse(inVector));
  endmethod
  
  // Interface method used to get the transformed data from the FFT
  // Module.  Examine AudioPipeline.bsv to see how this interface method
  // is used
  method ActionValue#(Vector#(`NUM_POINTS, Complex#(COMPLEX_DATA))) fftOutput(); 
    outputFIFO.deq();
    return outputFIFO.first(); 
  endmethod

endmodule

